/**********************
The Mandelbrot set

My learning curve just shot up - so I thought I'd start learning Mesa, GLUT, etc. with
this ever so wonderful fractal.
**********************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stdarg.h>
#include <GL/glut.h>

static void Key( unsigned char key, int x, int y );
static void Display(void);
static void Reshape(int newwidth, int newheight);
void calc_colour(int x, int y, float *colour);
static void Mouse(int button, int state, int mouseX, int mouseY);
void menufunc(int value);
void build_menu(void);
void clear_window(void);


#define INITIAL_ACCURACY 256

int Wwidth = 200, Wheight = 200;
float xoffset = -2, yoffset = -2, xrange = 4, yrange = 4;
int accuracy = INITIAL_ACCURACY, constrain = 1, maximized = 0;

int main(int argc, char *argv[]){

	glutInit( &argc, argv );
	//glutInitWindowPosition(0, 0);
	glutInitWindowSize(Wwidth, Wheight);

	glutInitDisplayMode(GLUT_RGB);// | GLUT_DOUBLE);

	if(glutCreateWindow("The Mandelbrot Set") == GL_FALSE) exit(1);
	glEnable(GL_DEPTH_TEST);
	glutReshapeFunc(Reshape);
	glutKeyboardFunc(Key);
	glutDisplayFunc(Display);
	glutMouseFunc(Mouse);
	glutCreateMenu(&menufunc);
	build_menu();
	glutAttachMenu(2);

	glutMainLoop();

	return 0;
}

void clear_window(void){
	/******** clear the viewing area ********/
	glBegin(GL_POLYGON);
		glColor3f(0, 0, 0);
		glVertex2f(0, 0);
		glVertex2f(0, Wheight + 1);
		glVertex2f(Wwidth + 1, Wheight + 1);
		glVertex2f(Wwidth + 1, 0);
	glEnd();
	glFlush();
}

static void Reshape(int newwidth, int newheight){
	Wwidth = newwidth;
	Wheight = newheight;
	glViewport(0, 0, Wwidth - 1, Wheight - 1);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0, (GLint)Wwidth, 0, (GLint)Wheight);
	glMatrixMode(GL_MODELVIEW);
	clear_window();
}

static void Display(void){
	float colour[3];
	int x, y;

	glutSetCursor(7);

//	clear_window();
//	glutSwapBuffers();

	/********* draw the fractal **********/
	glPointSize(1);
	glBegin(GL_POINTS);
		for(x = 0; x <= Wwidth; x++){
			for(y = 0; y <= Wheight; y++){
				calc_colour(x, y, colour);
				glColor3fv(colour);
				glVertex2f(x, y);
			}
		}
	glEnd();
	glFlush();
//	glutSwapBuffers();

	glutSetCursor(1);

}

static void Key( unsigned char key, int x, int y ){
	switch (key) {
		case 27:
			exit(0);
			break;
		case 'a':
			if(accuracy < 32768){
				accuracy <<= 1;
				Display();
			}
			break;
		case 'A':
			if(accuracy > 1){
				accuracy >>= 1;
				Display();
			}
			break;
		case 13:
			xoffset = yoffset = -2.;
			xrange = yrange = 4.;
			accuracy = INITIAL_ACCURACY;
			Display();
			break;
	}
}

void calc_colour(int x, int y, float *colour){
	double c = (xrange / (float)Wwidth) * (float)x + (float)xoffset;
	double ci = (yrange / (float)Wheight) * (float)y + (float)yoffset;
	int count = 0;
	double z = 0, zi = 0, zsq = 0;
	double zisq = 0, nz = 0, nzi = 0;

	while((count <= accuracy) && ((zsq + zisq) < 4.)){
		nz = zsq - zisq;
		nzi = (z * zi) * 2.;
		z = nz + c;
		zi = nzi + ci;
		zsq = z * z;
		zisq = zi * zi;
		count++;
	}

	colour[0] = .5 - cos((float)zsq / 10.) * .25 - sin((float)count / 3.) * .5;
	colour[1] = .5 - sin((float)zsq / 10.) * .25 + sin((float)count / 10.) * .5;
	colour[2] = .5 + cos((float)zsq / 10.) * .25 - cos((float)count / 7.) * .5;
/*
	colour[0] = .5 - sin((float)count) * .5;
	colour[2] = .5 - cos((float)count / 4.) * .5;
	colour[1] = .5 + sin((float)count / 2.) * .5;
*/
}

static void Mouse(int button, int state, int mouseX, int mouseY){	
	static int x1, y1, x2, y2, deltax, deltay;

	if(state == GLUT_DOWN){
		x1 = mouseX;
		y1 = Wheight - mouseY;
		glColor3f(1., 1., 1.);
		glBegin(GL_LINE_STRIP);
			glVertex3f(x1, y1 - 10, 0);		
			glVertex3f(x1, y1 + 10, 0);		
		glEnd();
		glBegin(GL_LINE_STRIP);
			glVertex3f(x1 - 10, y1, 0);		
			glVertex3f(x1 + 10, y1, 0);		
		glEnd();
		glPopMatrix();
		glFlush();
//		glutSwapBuffers();
	}else{
		x2 = mouseX;
		y2 = Wheight - mouseY;
		if(x1 > x2){x1 ^= x2; x2 ^= x1; x1 ^= x2; }
		if(y1 > y2){y1 ^= y2; y2 ^= y1; y1 ^= y2; }

		deltax = x2 - x1;
		deltay = y2 - y1;
		if(constrain){
			if(deltax != deltay){
				if(deltax < deltay){
					deltax = deltay;
					x2 = x1 + deltax;
				}else{
					deltay = deltax;
					y2 = y1 + deltay;
				}
			}
		}

		if(deltax && deltay){
			xoffset += ((double)x1 * xrange) / (double)Wwidth;
			yoffset += ((double)y1 * yrange) / (double)Wheight;
			xrange *= (double)(x2 - x1) / Wwidth;
			yrange *= (double)(y2 - y1) / Wheight;
			Display();
		}
	}
}

void menufunc(int value){
	switch(value){
		case 0:
			exit(0);
			break;
		case 1:
			xoffset = yoffset = -2.;
			xrange = yrange = 4.;
			accuracy = INITIAL_ACCURACY;
			Display();
			break;
		case 2:
			if(accuracy < 32768){
				accuracy <<= 1;
				Display();
			}
			break;
		case 3:
			if(accuracy > 1){
				accuracy >>= 1;
				Display();
			}
			break;
		case 4:
			constrain = !constrain;
			if(constrain)
				glutChangeToMenuEntry(5, "Constrain? (Yes)", 4);
			else
				glutChangeToMenuEntry(5, "Constrain? (No)", 4);
			break;
		case 5:
			if(maximized){
				Wwidth = Wheight = 200;
				glutReshapeWindow(Wwidth, Wheight);
				glutPositionWindow(100, 100);
				glutChangeToMenuEntry(6, "Maximize", 5);
			}else{
				glutChangeToMenuEntry(6, "Minimize", 5);
				glutFullScreen();
			}
			maximized = 1 - maximized;
			break;
	}
}

void build_menu(void){
	glutAddMenuEntry("Exit", 0);
	glutAddMenuEntry("Reset", 1);
	glutAddMenuEntry("Accuracy * 2", 2);
	glutAddMenuEntry("Accuracy / 2", 3);
	glutAddMenuEntry("Constrain? (Yes)", 4);
	glutAddMenuEntry("Maximize", 5);
}
